module I1M.Analizador where

import Data.Char

-- ---------------------------------------------------------------------
-- § El tipo de los analizadores                                      --
-- ---------------------------------------------------------------------

type Analizador a = String -> [(a,String)]      

analiza :: Analizador a -> String -> [(a,String)]
analiza a cs = a cs

-- ---------------------------------------------------------------------
-- § Analizadores básicos                                             --
-- ---------------------------------------------------------------------

-- analiza (resultado 3) "Hola"  ==  [(3,"Hola")]
resultado :: a -> Analizador a
resultado v =  \ent -> [(v,ent)]

-- analiza fallo "Hola"  ==  []
fallo :: Analizador a
fallo = \ent -> []

-- analiza elemento "Hola"  ==  [('H',"ola")]
-- analiza elemento ""      ==  []
elemento :: Analizador Char
elemento = \xs -> case xs of
                    [] -> []
                    (x:xs) -> [(x,xs)]

-- ---------------------------------------------------------------------
-- § Secuenciación                                                    --
-- ---------------------------------------------------------------------

-- ((p >*> f) e) falla si el análisis de e por p falla, en caso
-- contrario, se obtiene un valor (v) y una salida (s), se aplica la
-- función f al valor v obteniéndose un nuevo analizador con el que se
-- analiza la salida s.
infixr 5 >*>

(>*>) :: Analizador a -> (a -> Analizador b) -> Analizador b
p >*> f = \ent -> case analiza p ent of
                       []        -> []
                       [(v,sal)] -> analiza (f v) sal

-- primeroTercero es un analizador que devuelve la primero y tercer
-- carácter de la cadena. Por ejemplo, 
--    primeroTercero "Juan"  ==  [(('J','a'),"n")]
--    primeroTercero "Ju"    ==  []
primeroTercero :: Analizador (Char,Char)
primeroTercero = 
    elemento >*> \x ->
    elemento >*> \_ ->
    elemento >*> \y ->
    resultado (x,y)

-- ---------------------------------------------------------------------
-- § Elección                                                         --
-- ---------------------------------------------------------------------

-- ((p +++ q) e) analiza e con p y si falla analiza e con q. Por ejemplo,
--    analiza (elemento +++ resultado 'd') "abc"  ==  [('a',"bc")]
--    analiza (fallo +++ resultado 'd') "abc"     ==  [('d',"abc")]
--    analiza (fallo +++ fallo) "abc"             ==  []
(+++) :: Analizador a -> Analizador a -> Analizador a
p +++ q = \ent -> case analiza p ent of
                    []        -> analiza q ent
                    [(v,sal)] -> [(v,sal)]

-- ---------------------------------------------------------------------
-- Primitivas derivadas                                               --
-- ---------------------------------------------------------------------

-- (sat p) es el analizador que consume un elemento si dicho elemento
-- cumple la propiedad p y falla en caso contrario. Por ejemplo,
--    analiza (sat isLower) "hola"  ==  [('h',"ola")]
--    analiza (sat isLower) "Hola"  ==  []
sat :: (Char -> Bool) -> Analizador Char
sat p = elemento >*> \x ->
        if p x then resultado x else fallo

-- analiza digito "123"  ==  [('1',"23")]
-- analiza digito "uno"  ==  []
digito :: Analizador Char
digito = sat isDigit

-- analiza minuscula "eva"  ==  [('e',"va")]
-- analiza minuscula "Eva"  ==  []
minuscula :: Analizador Char
minuscula = sat isLower

-- analiza mayuscula "Eva"  ==  [('E',"va")]
-- analiza mayuscula "eva"  ==  []
mayuscula :: Analizador Char
mayuscula = sat isUpper

-- analiza letra "Eva"  ==  [('E',"va")]
-- analiza letra "eva"  ==  [('e',"va")]
-- analiza letra "123"  ==  []
letra :: Analizador Char
letra = sat isAlpha

-- analiza alfanumerico "Eva"   ==  [('E',"va")]
-- analiza alfanumerico "eva"   ==  [('e',"va")]
-- analiza alfanumerico "123"   ==  [('1',"23")]
-- analiza alfanumerico " 123"  ==  []
alfanumerico :: Analizador Char
alfanumerico = sat isAlphaNum

-- analiza (caracter 'E') "Eva"  ==  [('E',"va")]
-- analiza (caracter 'E') "eva"  ==  []
caracter :: Char -> Analizador Char
caracter x = sat (== x)

-- analiza (cadena "abc") "abcdef"  ==  [("abc","def")]
-- analiza (cadena "abc") "abdcef"  ==  []
cadena :: String -> Analizador String
cadena []     = resultado []
cadena (x:xs) = caracter x >*> \x  ->
                cadena xs  >*> \xs ->
                resultado (x:xs)

-- analiza (varios digito) "235abc"  ==  [("235","abc")]
-- analiza (varios digito) "abc235"  ==  [("","abc235")]
varios :: Analizador a -> Analizador [a]
varios p  = varios1 p +++ resultado []
                    
-- analiza (varios1 digito) "235abc"  ==  [("235","abc")]
-- analiza (varios1 digito) "abc235"  ==  []
varios1 :: Analizador a -> Analizador [a]
varios1 p = p        >*> \v  ->
            varios p >*> \vs ->
            resultado (v:vs)

-- analiza ident "lunes12 de Ene"  ==  [("lunes12"," de Ene")]
-- analiza ident "Lunes12 de Ene"  ==  []
ident :: Analizador String
ident =  minuscula           >*> \x  ->
         varios alfanumerico >*> \xs ->
         resultado (x:xs)

-- analiza nat "14DeAbril"   ==  [(14,"DeAbril")]
-- analiza nat " 14DeAbril"  ==  []
nat :: Analizador Int
nat = varios1 digito >*> \xs ->
      resultado (read xs)

-- analiza espacio "    a b c"  ==  [((),"a b c")]
espacio :: Analizador ()
espacio = varios (sat isSpace) >*> \_ -> 
          resultado ()

-- ---------------------------------------------------------------------
-- § Tratamiento de espacios                                          --
-- ---------------------------------------------------------------------

-- analiza (unidad nat) " 14DeAbril"     ==  [(14,"DeAbril")]
-- analiza (unidad nat) " 14   DeAbril"  ==  [(14,"DeAbril")]
unidad :: Analizador a -> Analizador a
unidad p = espacio >*> \_ ->
           p       >*> \v ->
           espacio >*> \_ ->
           resultado v

-- analiza identificador "  lunes12  de Ene"  ==  [("lunes12","de Ene")]
identificador :: Analizador String
identificador = unidad ident

-- analiza natural "  14DeAbril"  ==  [(14,"DeAbril")]
natural :: Analizador Int
natural =  unidad nat

-- analiza (simbolo "abc") "  abcdef"  ==  [("abc","def")]
simbolo :: String -> Analizador String
simbolo xs =  unidad (cadena xs)

-- analiza listaNat " [  2,  3, 5   ]"  ==  [([2,3,5],"")]
-- analiza listaNat " [  2,  3,]"       ==  []
listaNat :: Analizador [Int]
listaNat = simbolo "["          >*> \_ ->
           natural              >*> \n ->
           varios (simbolo ","  >*> \_ ->
                   natural)     >*> \ns ->
           simbolo "]"          >*> \_ ->
           resultado (n:ns)

-- ---------------------------------------------------------------------
-- § Expresiones aritméticas                                          --
-- ---------------------------------------------------------------------

-- analiza expr "2*3+5"     ==  [(11,"")]
-- analiza expr "2*(3+5)"   ==  [(16,"")]
-- analiza expr "2+3*5"     ==  [(17,"")]
-- analiza expr "2*3+5abc"  ==  [(11,"abc")]
expr :: Analizador Int
expr = term                  >*> \t ->
       (simbolo "+"          >*> \_ ->              
        expr                 >*> \e ->
        resultado (t+e))
       +++ (simbolo "-"      >*> \_ ->              
            expr             >*> \e ->
            resultado (t-e))
       +++ resultado t

-- analiza term "2*3+5"  ==  [(6,"+5")]
term :: Analizador Int
term = factor >*> \f ->
       (simbolo "*"                >*> \_ ->
        term                       >*> \t ->
        resultado (f*t))
       +++ (simbolo "/"            >*> \_ ->
            term                   >*> \t ->
            resultado (f `div` t))
       +++ resultado f

-- analiza factor "2*3+5"      ==  [(2,"*3+5")]
-- analiza factor "(2+3)*5"    ==  [(5,"*5")]
-- analiza factor "(2+3*7)*5"  ==  [(23,"*5")]
factor :: Analizador Int
factor =  (simbolo "(" >*> \_ ->
           expr        >*> \e ->
           simbolo ")" >*> \_ ->
           resultado e)
          +++ natural

-- valor "2*3+5"      ==  11
-- valor "2*(3+5)"    ==  16
-- valor "2 * 3 + 5"  ==  11
-- valor "2*3x+5y"    ==  *** Exception: entrada sin usar x+5y
-- valor "-1"         ==  *** Exception: entrada no valida
valor :: String -> Int
valor xs = case (analiza expr xs) of
             [(n,[])]  -> n
             [(_,sal)] -> error ("entrada sin usar " ++ sal)
             []        -> error "entrada no valida"

